@@ -16,24 +16,25 @@
 ;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 ;; Boston, MA 02111-1307, USA.
 
+;;; The loading of this file constitutes a START message as described
+;;; in EPL.pm.
 
 (require 'epl)
 
-(setq epl-interp
-      (vector 'perl-interpreter-tag
-	      nil ;in
-	      nil ;out
-	      (generate-new-buffer "perl") ;buffer
-	      (make-hash-table :test 'eq) ;gcpro
-	      1   ;next-handle
-	      (make-hash-table :test 'eq :weakness 'value) ;refs
-	      0   ;nrefs
-	      'running ;status
-	      ))
+(let* ((epl-interp (make-epl-interp :in nil :out nil :depth 0))
+       (perl-interpreter epl-interp)
+       (standard-input t))
+  (epl-puthash t epl-interp epl-interp-map)
+  ;; Check version
+  (if command-line-args-left
+      (let ((me (format "%d.%d" epl-major-version epl-minor-version))
+	    (you (car command-line-args-left)))
+	(if (string= you me)
+	    (setq command-line-args-left (cdr command-line-args-left))
+	  (epl-send-message "&cb_raise("
+			    (format "Version mismatch: %s versus epl.el %s"
+				    you me) ")"))))
+  ;; Answer the START and loop until told to exit.
+  (epl-send-and-receive "&cb_return()"))
 
-(setq epl-interp-map (make-hash-table :test 'eq))
-(puthash t epl-interp epl-interp-map)
-(add-hook 'kill-emacs-hook 'epl-kill-emacs-hook)
-(setq perl-interpreter epl-interp)
-
-(let ((standard-input t)) (epl-loop))
+(kill-emacs)
